# -*- coding: utf-8 -*-
import datetime
import os
from types import ListType

from plone.portlets.constants import CONTENT_TYPE_CATEGORY
from plone.portlets.constants import CONTEXT_CATEGORY
from plone.portlets.constants import GROUP_CATEGORY
from plone.portlets.interfaces import ILocalPortletAssignmentManager
from plone.portlets.interfaces import IPortletAssignmentMapping
from plone.portlets.interfaces import IPortletAssignmentSettings
from plone.portlets.interfaces import IPortletManager
from zope import event
from zope.component._api import getMultiAdapter
from zope.component._api import getUtility
from zope.container.interfaces import INameChooser

from DateTime.DateTime import DateTime
from Products.Archetypes.event import ObjectInitializedEvent
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.utils import safe_unicode
from zExceptions import BadRequest

# API
from wm.sampledata.images import getImage
from wm.sampledata.images import getFlickrImage
from wm.sampledata.images import getRandomImage
from wm.sampledata.images import getRandomFlickrImage
from wm.sampledata.images import get_placeholder_image

# Silence flake8
assert getImage
assert getFlickrImage
assert getRandomImage
assert getRandomFlickrImage
assert get_placeholder_image


IPSUM_LINE = "Lorem ipsum mel augue antiopam te. Invidunt constituto accommodare ius cu. Et cum solum liber doming, mel eu quem modus, sea probo putant ex."  # noqa

IPSUM_PARAGRAPH = "<p>" + 10 * IPSUM_LINE + "</p>"


def getFile(module, *path):
    """return the file located in module.
    if module is None, treat path as absolut path
    path can be ['directories','and','file.txt'] or just 'file.txt'
    """
    modPath = ''
    if module:
        modPath = os.path.dirname(module.__file__)

    if type(path) == str:
        path = [path]
    filePath = os.path.join(modPath, *path)
    return file(filePath)


def getFileContent(module, *path):
    f = getFile(module, *path)
    data = safe_unicode(f.read())
    f.close()
    return data


def deleteItems(folder, *ids):
    """delete items in a folder and don't complain if they do not exist.
    """
    for itemId in ids:
        try:
            folder.manage_delObjects([itemId])
        except BadRequest:
            pass
        except AttributeError:
            pass


def todayPlusDays(nrDays=0, zopeDateTime=False):
    today = datetime.date.today()
    date = today + datetime.timedelta(days=nrDays)
    if zopeDateTime:
        return DateTime(date.year, date.month, date.day)
    else:
        return date


def eventAndReindex(*objects):
    """fires an objectinitialized event and
    reindexes the object(s) after creation so it can be found in the catalog
    """
    for obj in objects:
        event.notify(ObjectInitializedEvent(obj))
        obj.reindexObject()


def workflowAds(home, wfdefs):
    """
    do workflow transitions and set enddate to datetime if set.

    sample format
    wfdefs = [('plone-dev', ['publish'], None),
              ('minimal-job', ['submit'], datetime),
              ('plone-dev', ['publish']),
              ]
    """

    wft = getToolByName(home, 'portal_workflow')

    for id, actions, date in wfdefs:
        ad = home.unrestrictedTraverse(id)
        for action in actions:
            wft.doActionFor(ad, action)
        if date:
            ad.expirationDate = date
        ad.reindexObject(idxs=['end', 'review_state'])


def addPortlet(context, columnName='plone.leftcolumn', assignment=None):
    if not assignment:
        return
    column = getUtility(IPortletManager, columnName)
    manager = getMultiAdapter((context, column), IPortletAssignmentMapping)
    chooser = INameChooser(manager)
    manager[chooser.chooseName(None, assignment)] = assignment


def removePortlet(context, portletName, columnName='plone.leftcolumn'):
    manager = getUtility(IPortletManager, columnName)
    assignmentMapping = getMultiAdapter(
        (context, manager), IPortletAssignmentMapping)
    # throws a keyerror if the portlet does not exist
    del assignmentMapping[portletName]


def blockPortlets(context, columnName='plone.leftcolumn',
                  inherited=None, group=None, contenttype=None):
    """True will block portlets, False will show them, None will skip settings.
    """

    manager = getUtility(IPortletManager, name=columnName)
    assignable = getMultiAdapter(
        (context, manager), ILocalPortletAssignmentManager)

    if group is not None:
        assignable.setBlacklistStatus(GROUP_CATEGORY, group)
    if contenttype is not None:
        assignable.setBlacklistStatus(CONTENT_TYPE_CATEGORY, contenttype)
    if inherited is not None:
        assignable.setBlacklistStatus(CONTEXT_CATEGORY, inherited)


def hidePortlet(context, portletName, columnName='plone.leftcolumn'):
    manager = getUtility(IPortletManager, columnName)
    assignmentMapping = getMultiAdapter(
        (context, manager), IPortletAssignmentMapping)
    settings = IPortletAssignmentSettings(assignmentMapping[portletName])
    settings['visible'] = False


def hasPortlet(context, portletName, columnName='plone.leftcolumn'):
    manager = getUtility(IPortletManager, columnName)
    assignmentMapping = getMultiAdapter(
        (context, manager), IPortletAssignmentMapping)
    return portletName in assignmentMapping


def setPortletWeight(portlet, weight):
    """if collective weightedportlets can be imported
    we do set the weight, and do not do anything otherwise
    """
    try:
        from collective.weightedportlets import ATTR
        from persistent.dict import PersistentDict
        if not hasattr(portlet, ATTR):
            setattr(portlet, ATTR, PersistentDict())
        getattr(portlet, ATTR)['weight'] = weight
    except ImportError:
        # simply don't do anything in here
        pass


def createImage(context, id, file, title='', description=''):
    """create an image and return the object
    """
    context.invokeFactory('Image', id, title=title,
                          description=description)
    context[id].setImage(file)
    return context[id]


def createFile(context, id, file, title='', description=''):
    context.invokeFactory('File', id, title=title,
                          description=description)
    context[id].setFile(file)
    return context[id]


def excludeFromNavigation(obj, exclude=True):
    """excludes the given obj from navigation
    make sure to reindex the object afterwards to make the
    navigation portlet notice the change
    """

    obj._md['excludeFromNav'] = exclude


def getRelativePortalPath(context):
    """return the path of the plonesite
    """
    url = getToolByName(context, 'portal_url')
    return url.getPortalPath()


def getRelativeContentPath(obj):
    """return the path of the object
    """
    url = getToolByName(obj, 'portal_url')
    return '/'.join(url.getRelativeContentPath(obj))


def doWorkflowTransition(obj, transition):
    """to the workflow transition on the specified object
    we don't use wft.doActionFor directly since this does not set the effective
    data
    """

    doWorkflowTransitions([obj], transition)


def doWorkflowTransitions(objects=[], transition='publish',
                          includeChildren=False):
    """use this to publish a/some folder(s) optionally
    including their child elements
    """

    if not objects:
        return
    if type(objects) != ListType:
        objects = [objects, ]

    utils = getToolByName(objects[0], 'plone_utils')
    for obj in objects:
        path = '/'.join(obj.getPhysicalPath())
        utils.transitionObjectsByPaths(
            workflow_action=transition,
            paths=[path],
            include_children=includeChildren
        )


def constrainTypes(obj, allowed=[], notImmediate=[]):
    """sets allowed and immediately addable types for obj.

    to only allow news and images and make both immediately addable use::

       constrainTypes(portal.newsfolder, ['News Item', 'Image'])

    if images should not be immediately addable you would use::

       constrainTypes(portal.newsfolder, ['News Item', 'Image'],
                      notImmediate=['Image'])
    """

    obj.setConstrainTypesMode(1)
    obj.setLocallyAllowedTypes(allowed)

    if notImmediate:
        immediate = [type for type in allowed if type not in notImmediate]
    else:
        immediate = allowed
    obj.setImmediatelyAddableTypes(immediate)


def raptus_hide_for(item, component):
    """hide the specified item in the `raptus.article` component given by it's name
    (eg. ``(item=<Image>, component='imageslider.teaser')`` )
    """
    components = list(item.Schema()['components'].get(item))
    item.Schema()['components'].set(
        item, [c for c in components if not c == component])
    item.reindexObject()


def raptus_show_for(item, component):
    """show the specified item in the `raptus.article` component given by it's name
    (eg. ``(item=<Image>, component='imageslider.teaser')`` )
    """
    components = list(item.Schema()['components'].get(item))
    item.Schema()['components'].set(
        item, [c for c in components if not c == component])
    item.reindexObject()



# -*- coding: utf-8 -*-
import os
import unittest

import wm.sampledata.tests
from wm.sampledata import example
from wm.sampledata.utils import getFileContent


class FileUtilities(unittest.TestCase):

    def testPathWithModule(self):
        fc = getFileContent(wm.sampledata.tests, 'example', 'portlet.html')
        self.assertEqual(len(fc), 230)

    def testFileWithModule(self):
        fc = getFileContent(example, 'portlet.html')
        self.assertEqual(len(fc), 230)

    def testFileWithoutModule(self):
        pkgdir = os.path.dirname(wm.sampledata.tests.__file__)
        exdir = os.path.join(pkgdir, 'example')

        fc = getFileContent(None, exdir, 'portlet.html')
        self.assertEqual(len(fc), 230)

    def testNoFileWithModule(self):
        self.assertRaises(IOError, getFileContent, example, 'notexistent.html')


def test_suite():
    s = unittest.TestSuite()
    s.addTest(unittest.makeSuite(FileUtilities))
    return s
#
# def main():
#    unittest.TextTestRunner().run(test_suite())
#
# if __name__=='__main__':
#    if len(sys.argv) > 1:
#        globals()[sys.argv[1]]()
#    else:
#        main()



# -*- coding: utf-8 -*-
from functools import wraps
from wm.sampledata import logger
from PIL import Image
from io import BytesIO


import os
import random
import time
import urllib
import requests



# see
# http://www.saltycrane.com/blog/2009/11/trying-out-retry-decorator-python/
# http://stackoverflow.com/questions/9446387/how-to-retry-urllib2-request-when-failed
def _retry(ExceptionToCheck, tries=4, delay=3, backoff=2, logger=None):
    """Retry calling the decorated function using an exponential backoff.

    http://www.saltycrane.com/blog/2009/11/trying-out-retry-decorator-python/
    original from: http://wiki.python.org/moin/PythonDecoratorLibrary#Retry

    :param ExceptionToCheck: the exception to check. may be a tuple of
        exceptions to check
    :type ExceptionToCheck: Exception or tuple
    :param tries: number of times to try (not retry) before giving up
    :type tries: int
    :param delay: initial delay between retries in seconds
    :type delay: int
    :param backoff: backoff multiplier e.g. value of 2 will double the delay
        each retry
    :type backoff: int
    :param logger: logger to use. If None, print
    :type logger: logging.Logger instance
    """
    def deco_retry(f):

        @wraps(f)
        def f_retry(*args, **kwargs):
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except ExceptionToCheck, e:
                    msg = "%s, Retrying in %d seconds..." % (str(e), mdelay)
                    if logger:
                        logger.warning(msg)
                    else:
                        print msg
                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff
            return f(*args, **kwargs)

        return f_retry  # true decorator

    return deco_retry


@_retry(requests.RequestException, tries=5, delay=2, logger=logger)
def _download(url):
    """try to download the url 5 times, with exponential delay.
    first delay 2 seconds, second 4, third 8, etc
    """
    logger.info('Downloading %s' % url)
    headers = {'User-Agent': 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:46.0) Gecko/20100101 Firefox/46.0'}  # noqa
    return requests.get(url, headers=headers).content


# Lorempixel.com categories.
CATEGORIES = [
    'abstract', 'animals', 'business',
    'cats', 'city', 'food', 'nightlife',
    'fashion', 'people', 'nature', 'sports',
    'technics', 'transport',
]


def get_placeholder_image(width=1024, height=768,
                          text='',  # default to dimensions of image
                          bkg_color='CCCCCC',
                          text_color='969696'):
    """see https://placeholder.com/ for explanations
    """

    url = u'http://via.placeholder.com/{w}x{h}/{bkg}/{textcolor}'.format(
        w=width,
        h=height,
        bkg=bkg_color.strip('#'),
        textcolor=text_color.strip('#'))

    params = {}
    if text:
        params['text'] = text.encode('utf-8')

    if params:
        url += '?' + urllib.urlencode(params)

    return _download(url)


def test_stub(f):
    """Checks for environment variable "CI=true" and return a  grey jpeg
    generated with pil instead with dimensions width and height
    """

    @wraps(f)
    def f_stub(*args, **kwargs):
        width = 1024
        height = 768
        if len(args) >= 2:
            width, height = args[:2]
        ci = os.environ.get('CI', '0')
        if ci.lower() in ['1', 'true', 'yes', True]:
            logger.info("CI=true set, do not download image")
            img = Image.new("RGB", (width, height), '#bababa')
            with BytesIO() as buffer:
                img.save(buffer, 'JPEG')
                buffer.seek(0)
                return buffer.read()
        return f(*args, **kwargs)

    return f_stub  # true decorator


@test_stub
def getFlickrImage(width=1024, height=768,
                   keywords=[], match_all_keywords=False,
                   gray=False):
    """obtains an image from loremflickr.com.

    If you set match_all_keywords to true it will
    search for all keywords (AND).

    Set gray to True to get a grayscale image.
    """

    url = 'http://loremflickr.com/'
    params = []

    if gray:
        params.append('g')

    params.append('%d/%d' % (width, height))
    params.append(','.join(keywords))
    if match_all_keywords:
        params.append('all')

    url = url + '/'.join(params)

    return _download(url)


@test_stub
def getImage(width=1024, height=768, category=None,
             gray=False, index=None, text=None):
    """obtains an image from lorempixel.com

    for possible categories see ``CATEGORIES``.
    """

    url = 'http://lorempixel.com/'
    params = []

    if gray:
        params.append('g/')
    params.append('%d/%d/' % (width, height))
    if category:
        params.append(category + '/')
        if index:
            params.append('%d/' % index)
    if text:
        params.append('%s/' % text)

    url = url + ''.join(params)

    return _download(url)


RATIOS = [(16, 9), (4, 3)]


def getRandomFlickrImage(long_edge=1024,
                         keywords=[], match_all_keywords=False,
                         gray=None, landscape=None, ratios=RATIOS):
    """Returns a random image from loremflickr.com (portrait or landscape)
    in one of the available RATIOS.

    Set gray to True to force grayscale pictures and to False to force color
    pictures.
    """
    ratio = random.choice(ratios)
    ratio = float(ratio[0]) / ratio[1]

    if landscape is None:
        landscape = random.choice([True, False])

    if gray is None:
        gray = random.choice([True, False])

    if landscape:
        width = long_edge
        height = int(long_edge / ratio)
    else:
        height = long_edge
        width = int(long_edge / ratio)

    return getFlickrImage(
        width, height,
        keywords=keywords,
        match_all_keywords=match_all_keywords,
        gray=gray
    )


def getRandomImage(long_edge=1024, category=None, gray=None,
                   landscape=None, ratios=RATIOS):
    """returns a random image from lorempixel.com (portrait or landscape)
    in one of the available RATIOS

    set gray to True to force grayscale pictures and to False to force color
    pictures
    """
    ratio = random.choice(ratios)
    ratio = float(ratio[0]) / ratio[1]

    if landscape is None:
        landscape = random.choice([True, False])

    if gray is None:
        gray = random.choice([True, False])

    if landscape:
        width = long_edge
        height = int(long_edge / ratio)
    else:
        height = long_edge
        width = int(long_edge / ratio)

    return getImage(width, height, category=category, gray=gray)



# -*- coding: utf-8 -*-
from zope.interface import Interface
from zope.interface.interface import Attribute


class ISampleDataPlugin(Interface):

    """
    """

    title = Attribute(u"Plugin Title")

    description = Attribute(u"Describes what is being generated")

    def generate(context):
        """Generate sample data of this plugin.
        """

        pass



# -*- coding: utf-8 -*-
import logging

from zope.component import getUtility
from zope.interface.declarations import implements

from wm.sampledata.interfaces import ISampleDataPlugin


logger = logging.getLogger('wm.sampledata')


class PluginGroup(object):
    """useful baseclass for grouping plugins by their name
    """

    implements(ISampleDataPlugin)

    PLUGINS = []

    def generate(self, context):
        for plugin in self.PLUGINS:
            if isinstance(plugin, basestring):
                plugin = getUtility(ISampleDataPlugin, name=plugin)
                plugin.generate(context)
            else:
                plugin().generate(context)



# -*- coding: utf-8 -*-
from operator import itemgetter

from zope.component import getUtilitiesFor
from zope.component import getUtility
from zope.component.interfaces import ComponentLookupError

from Products.Five.browser import BrowserView
from Products.statusmessages.interfaces import IStatusMessage
from wm.sampledata import logger
from wm.sampledata.interfaces import ISampleDataPlugin


class SampleDataView(BrowserView):

    def listPlugins(self):
        """list all available plugins sorted by their name
        """
        plugins = []
        for name, util in getUtilitiesFor(ISampleDataPlugin):
            plugins.append(dict(name=name,
                                title=util.title,
                                description=util.description))

        return sorted(plugins, key=itemgetter('name'))

    def runPlugin(self, plugin, debug=False):
        """run a named plugin and redirect to the sampledata page again.
        show a status message that tells the user if the plugin
        could not be found, raised an error or ran successfully.
        """
        result = None
        try:
            plugin = getUtility(ISampleDataPlugin, name=plugin)
            result = plugin.generate(self.context)
            IStatusMessage(self.request).addStatusMessage(
                u"successfully ran %s" % plugin.title, 'info')
        except ComponentLookupError, e:
            IStatusMessage(self.request).addStatusMessage(
                u"could not find plugin %s" % plugin, 'error')
            logger.exception("could not find plugin %s" % plugin)

            if debug:
                raise e
        except Exception, e:
            IStatusMessage(self.request).addStatusMessage(
                u"error running %s: %s" % (plugin.title, str(e)), 'error')
            logger.exception(
                ("error running %s - try @@sampledata/run?"
                 "plugin=<plugin-name>&debug=True") % (plugin.title))

            if debug:
                raise e
        finally:
            if not debug:
                # return to listing
                self.request.response.redirect(
                    self.context.absolute_url() + '/@@' + self.__name__)
                return result



# -*- coding: utf-8 -*-
from plone.portlet.static.static import Assignment as StaticAssignment
from zope.interface.declarations import implements

from wm.sampledata.interfaces import ISampleDataPlugin
from wm.sampledata.utils import addPortlet
from wm.sampledata.utils import createImage
from wm.sampledata.utils import deleteItems
from wm.sampledata.utils import doWorkflowTransitions
from wm.sampledata.utils import eventAndReindex
from wm.sampledata.utils import getFileContent
from wm.sampledata.utils import getRandomImage
from wm.sampledata.utils import IPSUM_PARAGRAPH


class DemoContent(object):
    implements(ISampleDataPlugin)

    title = u"Demo Content"

    description = u"""Creates a document with an image and assigns a portlet
that displays contact information."""

    def generate(self, context):
        pageId = 'sample'

        # delete sample document if it exists
        deleteItems(context, pageId)

        context.invokeFactory('Document', id=pageId, title=u"Sample Document")
        page = context[pageId]

        # download image from lorempixel.com - force colour images from
        # category nature
        imageId = 'sampleImage'
        deleteItems(context, imageId)
        createImage(context, imageId, +
                    getRandomImage(category='nature', gray=False),
                    title=u"Random Image",
                    description=u"Downloaded from lorempixel.com")

        text = '<img class="image-right" src="%s/@@images/image/mini" />' % (
            imageId) + IPSUM_PARAGRAPH

        # mimetype makes tiny recognize the text as HTML
        page.setText(text, mimetype='text/html')

        # publish and reindex (needed to make it show up in the navigation) the
        # page
        doWorkflowTransitions(page)
        eventAndReindex(page)

        import wm.sampledata.example as myModule
        portlet = StaticAssignment(
            u"Contact", getFileContent(myModule, 'portlet.html'))
        addPortlet(page, 'plone.leftcolumn', portlet)



from setuptools import find_packages
from setuptools import setup


version = '0.7.0'

long_description = (
    open("README.rst").read() + "\n\n" +
    open("CHANGES.rst").read() + "\n"
    )

setup(name='wm.sampledata',
      version=version,
      description="UI and utility methods to generate sampledata for Plone projects",
      long_description=long_description,
      # Get more strings from
      # http://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
        "Programming Language :: Python",
        'Framework :: Plone',
        'Framework :: Plone :: 4.1',
        'Framework :: Plone :: 4.2',
        'Framework :: Plone :: 4.3',
        ],
      keywords='plone sampledata generation',
      author='Harald Friessnegger',
      author_email='office@lovelysystems.com',
      url='https://github.com/collective/wm.sampledata',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['wm'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          'requests',
      ],
      entry_points="""
      # -*- Entry points: -*-

      [z3c.autoinclude.plugin]
      target = plone
      """,
      )
